@@ -23,7 +23,7 @@
  * along with this program; if not, write to the Free Software
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  *
- * $Id: mfcr2.h,v 1.15 2005/01/28 23:29:45 steveu Exp $
+ * $Id: mfcr2.h,v 1.24 2005/06/21 14:36:33 steveu Exp $
  */
 
 #if !defined(_MFCR2_H_)
@@ -82,6 +82,40 @@
    variants is added.
  */
 
+/*
+    Some China specific signals
+    ---------------------------
+
+    Calling party category group I (KA) signals
+
+         Step by step switch           Crossbar or electronic switch
+         -------------------           -----------------------------
+    1  - regular, fixed delay          voice/fax/data, fixed delay
+    2  - regular, immediate, meter     voice/fax/data, immediate, meter
+    3  - regular, business             voice/fax/data, business
+    4  - priority #1                   voice/fax/data, priority #1
+    5  - free                          free
+    6  - small PBX                     small PBX
+    7  - priority #1, fixed delay      voice/fax/data, priority #1, fixed delay
+    8  - priority #2, fixed delay      voice/fax/data, priority #2, fixed delay
+    9  - privileged local              voice/fax/data, priority #1, business
+    10 - non-privileged local          free
+    11 - spare                         spare
+    12 - spare                         spare
+    13 - test                          test
+    14 - spare                         spare
+    15 - spare                         spare
+
+    Calling party category group II (KD) signals (typical values)
+    
+    1  - Long distance operator, semi-automatic
+    2  - Long distance, automatic switching
+    3  - Local call, voice
+    4  - Local call, fax and data
+    5  - Semi-automatic, verify calling party's number
+    6  - Test call
+*/
+
 /* There are only 15 real signals, but the ITU specs. give them different
    names in different contexts, so lets do the same. That was there is some
    correlation with the specs. */
@@ -121,17 +155,18 @@
                             /* Other: End of pulsing (code 15) */
 
 /* R2 group II forward signals */
-#define R2_SIGII_1  '1'     /* National use - subscriber initiated call */
-#define R2_SIGII_2  '2'     /* National use - priority call */
-#define R2_SIGII_3  '3'     /* National use - maintenance equipment call */
-#define R2_SIGII_4  '4'     /* National use - spare */
-#define R2_SIGII_5  '5'     /* National use - operator initiated call */
-#define R2_SIGII_6  '6'     /* National use - data transmission call */
-#define R2_SIGII_7  '7'     /* Subscriber (or operator without forward
-                               transfer facility) */
-#define R2_SIGII_8  '8'     /* Data transmission call */
-#define R2_SIGII_9  '9'     /* Subscriber with priority */
-#define R2_SIGII_10 '0'     /* Operator with forward transfer facility */
+#define R2_SIGII_1  '1'     /* National - subscriber initiated call */
+#define R2_SIGII_2  '2'     /* National - subscriber with priority call */
+#define R2_SIGII_3  '3'     /* National - maintenance equipment call */
+#define R2_SIGII_4  '4'     /* National - spare */
+#define R2_SIGII_5  '5'     /* National - operator initiated call */
+#define R2_SIGII_6  '6'     /* National - data transmission call */
+#define R2_SIGII_7  '7'     /* International - Subscriber (or operator without
+                               forward transfer facility) */
+#define R2_SIGII_8  '8'     /* International - data transmission call */
+#define R2_SIGII_9  '9'     /* International - subscriber with priority */
+#define R2_SIGII_10 '0'     /* International - operator with forward transfer
+                               facility */
 #define R2_SIGII_11 'A'     /* Spare for national use */
 #define R2_SIGII_12 'B'     /* Spare for national use */
 #define R2_SIGII_13 'C'     /* Spare for national use */
@@ -157,148 +192,105 @@
 
 /* R2 group B backward signals */
 #define R2_SIGB_1   '1'     /* Spare for national use */
-#define R2_SIGB_2   '2'     /* Subscriber transferred */
+#define R2_SIGB_2   '2'     /* Subscriber transferred - send SIT to the caller */
 #define R2_SIGB_3   '3'     /* Subscriber line busy */
 #define R2_SIGB_4   '4'     /* Congestion (encountered after change from
                                group A to group B). */
-#define R2_SIGB_5   '5'     /* Vacant national number */
+#define R2_SIGB_5   '5'     /* Unassigned number */
 #define R2_SIGB_6   '6'     /* Subscriber line free - charge */
 #define R2_SIGB_7   '7'     /* Subscriber line free - no charge */
 #define R2_SIGB_8   '8'     /* Subscriber line out of order */
 #define R2_SIGB_9   '9'     /* Spare for national use */
 #define R2_SIGB_10  '0'     /* Spare for national use */
-#define R2_SIGB_11  'A'     /* Spare for international use */
-#define R2_SIGB_12  'B'     /* Spare for international use */
-#define R2_SIGB_13  'C'     /* Spare for international use */
-#define R2_SIGB_14  'D'     /* Spare for international use */
-#define R2_SIGB_15  'E'     /* Spare for international use */
-
-/* National variants:
-
-    Argentina uses signals reserved for national use as follows:
-    
-    R2_SIGA_9       Send the last digit again
-    R2_SIGA_10      Repeat the DID digits from the beginning
-    R2_SIGB_2       Send special information tone, and setup speech conditions
-    R2_SIGI_12      ANI not available (if no ANI digits yet received)
-    R2_SIGI_12      End of ANI, presentation restricted
-    R2_SIGI_15      End of ANI, presentation allowed
-    
-    Bahrain is as the ITU spec.
-
-    Bolivia uses signals reserved for national use as follows:
-
-    R2_SIGA_9       Repeat the DID digits from the beginning
-    R2_SIGB_8       Not used
-    R2_SIGI_15      End of DID digits
-    R2_SIGI_15      End of ANI digits
-    R2_SIGI_12      Category not available
-    R2_SIGI_12      ANI not available
-
-    Brazil uses signals reserved for national use as follows:
-
-    R2_SIGA_2       Repeat the DID digits from the beginning
-    R2_SIGA_6       Not used
-    R2_SIGA_9       Send last digit (n - 1)
-    R2_SIGB_1       Subscriber line free, charge.
-    R2_SIGB_2       Subscriber line busy
-    R2_SIGB_5       Subscriber line free, no charge.
-    R2_SIGB_6       Not used
-    R2_SIGB_7       Unassigned number
-    R2_SIGB_15      Not used
-    R2_SIGI_12      Category not available
-    R2_SIGI_12      ANI not available
-    R2_SIGI_15      End of ANI
-
-    Chile is as the ITU spec, except the C bit is used for metering pulses.
-    
-    China uses signals in the following non-ITU ways:
-
-    R2_SIGA_1       Send next digit (n + 1) - no matter whether DID or ANI
-    R2_SIGA_2       Repeat the DID digits from the beginning
-    R2_SIGA_5       Not used
-    R2_SIGA_6       Send calling party's category (followed by ANI)
-    R2_SIGA_7       Not used
-    R2_SIGA_8       Not used
-    R2_SIGA_9       Not used
-    R2_SIGA_10      Not used
-    R2_SIGB_1       Subscriber line free, charge.
-    R2_SIGB_2       Subscriber line busy
-    R2_SIGB_3       Not used
-    R2_SIGB_7       Not used
-
-    R2_SIGI_15      End of ANI
-    R2_SIGI_15      Category not available
-    R2_SIGI_15      ANI not available
-    
-    Honduras is as the ITU spec.
-
-    Korea uses signals reserved for national use as follows:
-
-    R2_SIGA_9       Repeat the DID digits from the beginning
-    
-    Panama uses signals reserved for national use as follows:
-
-    R2_SIGA_9       Repeat the DID digits from the beginning
-  
-    Thailand uses exactly the same variant as China.
- */
-
-/* Now some definitions for the states of the R2 protocol level state machine */
-
-enum
-{
-    /* Group A states for R2 protocol inbound (backward) */
-    R2_SENT_DNIS_REQUEST                    = 101,    /* Asking for next DNIS digit */
-    R2_SENT_DNIS_REQUEST_AFTER_ANI_DONE     = 102,    /* Asking for next DNIS digit after ANI complete */
-    R2_SENT_ANI_REQUEST                     = 103,    /* Asking for next ANI digit */
-    R2_SENT_ANI_REQUEST_BEFORE_DNIS_DONE    = 104,    /* Asking for next ANI digit before DNIS complete */
-    R2_SENT_CATEGORY_REQUEST                = 105,    /* Asking for calling party's category */
-    R2_SENT_CHANGE_TO_GROUP_II              = 106,    /* Asking for change to group II */
-
-    /* Group B states for R2 protocol inbound (backward) */
-    R2_SENT_ACCEPTED_CHARGE                 = 107,    /* Sent numbers complete and start ANSWER. Charge */
-    R2_SENT_ACCEPTED_NO_CHARGE              = 108,    /* Sent numbers complete and start ANSWER. No charge */
-    R2_SENT_NETWORK_CONGESTION              = 109,
-    R2_SENT_UNASSIGNED_NUMBER               = 110,
-    R2_SENT_USER_BUSY                       = 111,
-    R2_SENT_DEST_OUT_OF_ORDER               = 112,
-
-    /* Group I states for R2 protocol outbound (forward) */
-    R2_SENT_DNIS                            = 201,
-    R2_SENT_DNIS_AFTER_ANI_DONE             = 202,
-    R2_SENT_ANI                             = 203,
-    R2_SENT_ANI_BEFORE_DNIS_DONE            = 204,
-    R2_SENT_END_OF_DNIS                     = 205,
-    R2_SENT_END_OF_DNIS_AFTER_ANI_DONE      = 206,
-    R2_SENT_END_OF_ANI                      = 207,
-    R2_SENT_END_OF_ANI_BEFORE_DNIS_DONE     = 208
-};
+#define R2_SIGB_11  'A'     /* Spare for national use */
+#define R2_SIGB_12  'B'     /* Spare for national use */
+#define R2_SIGB_13  'C'     /* Spare for national use */
+#define R2_SIGB_14  'D'     /* Spare for national use */
+#define R2_SIGB_15  'E'     /* Spare for national use */
 
 enum
 {
     /* Call states for R2 protocol inbound (backward) */
-    R2_BACK_SEIZE_ACK                       = 101,    /* Sending SEIZE ACK in response to SEIZE */
-    R2_BACK_GROUP_A                         = 102,    /* Proceeding with group A signals */
-    R2_BACK_GROUP_B                         = 103,    /* Proceeding with group B signals */
-    R2_BACK_ANSWER                          = 104,    /* Sending ANSWER signal, after receiving SIGII_1 tone off */
-    R2_BACK_CLEAR_BACK                      = 105,    /* Sending CLEAR BACK after ANSWER */
-    R2_BACK_CLEAR_FWD                       = 106,    /* Got CLEAR FWD before CLEAR BACK */
-    R2_BACK_RELEASE_GUARD                   = 107,    /* Sending timed RELEASE GUARD to prevent early re-seize from our side. */
+    /*! Sending SEIZE ACK in response to SEIZE */
+    R2_BACK_SEIZE_ACK                           = 101,
+    /*! Sending group A signals */
+    R2_BACK_GROUP_A                             = 102,
+    /*! Sending group B signals */
+    R2_BACK_GROUP_B                             = 103,
+    /*! Sending group C signals (only used in a couple of countries) */
+    R2_BACK_GROUP_C                             = 104,
+    /*! Sending ANSWER signal, after receiving SIGII_1 tone off */
+    R2_BACK_ANSWER                              = 105,
+    /*! Sending CLEAR BACK after ANSWER */
+    R2_BACK_CLEAR_BACK                          = 106,
+    /*! Got CLEAR FWD before CLEAR BACK */
+    R2_BACK_CLEAR_FWD                           = 107,
+    /*! Sending timed RELEASE GUARD to prevent early re-seize from our side. */
+    R2_BACK_RELEASE_GUARD                       = 108,
 
     /* Call states for R2 protocol outbound (forward) */
-    R2_FWD_SEIZE                            = 201,    /* Sending SEIZE */
-    R2_FWD_GROUP_I                          = 202,    /* Proceeding with group I signals */
-    R2_FWD_GROUP_II                         = 203,    /* Proceeding with group II signals */
-    R2_FWD_WAIT_FOR_ANSWER                  = 204,    /* Waiting for CAS ANSWER after group B signal received */
-    R2_FWD_ANSWERED_BEFORE_GROUP_B_SIGNAL   = 205,    /* CAS ANSWER has preceeded group B signal */
-    R2_FWD_ANSWERED                         = 206,    /* The call is fully answered (group B tone + CAS bits) */
-    R2_FWD_CLEAR_BACK                       = 207,    /* CLEAR BACK after ANSWER */
-    R2_FWD_CLEAR_FWD_BEFORE_SEIZE_ACK       = 208,    /* Want to send CLEAR FWD before receiving SEIZE ACK */
-    R2_FWD_CLEAR_FWD_BEFORE_ANSWER          = 209,    /* Sending CLEAR FWD before receiving ANSWER */
-    R2_FWD_CLEAR_FWD_AFTER_ANSWER           = 210,    /* Sending CLEAR FWD after receiving ANSWER */
-    R2_FWD_CLEAR_FWD_AFTER_CLEAR_BACK       = 211     /* Sending CLEAR FWD after receiving CLEAR BACK. Waiting for RELEASE GUARD */
-};
+    /*! Sending SEIZE */
+    R2_FWD_SEIZE                                = 201,
+    /*! Proceeding with group I signals */
+    R2_FWD_GROUP_I                              = 202,
+    /*! Proceeding with group II signals */
+    R2_FWD_GROUP_II                             = 203,
+    /*! Proceeding with group III signals (only used in a couple of countries) */
+    R2_FWD_GROUP_III                            = 204,
+    /*! Waiting for CAS ANSWER after group B signal received */
+    R2_FWD_AWAIT_ANSWER                         = 205,
+    /*! CAS ANSWER has preceeded group B signal */
+    R2_FWD_ANSWERED_BEFORE_GROUP_B_SIGNAL       = 206,
+    /*! The call is fully answered (group B tone + CAS bits) */
+    R2_FWD_ANSWERED                             = 207,
+    /*! CLEAR BACK after ANSWER */
+    R2_FWD_CLEAR_BACK                           = 208,
+    /*! Want to send CLEAR FWD before receiving SEIZE ACK */
+    R2_FWD_CLEAR_FWD_BEFORE_SEIZE_ACK           = 209,
+    /*! Sending CLEAR FWD before receiving ANSWER */
+    R2_FWD_CLEAR_FWD_BEFORE_ANSWER              = 210,
+    /*! Sending CLEAR FWD after receiving ANSWER */
+    R2_FWD_CLEAR_FWD_AFTER_ANSWER               = 211,
+    /*! Sending CLEAR FWD after receiving CLEAR BACK. Waiting for RELEASE GUARD */
+    R2_FWD_CLEAR_FWD_AFTER_CLEAR_BACK           = 212
+} r2_phase_e;
+
+/* Now some definitions for the states of the R2 protocol level state machine */
+enum
+{
+    /* Group A states for R2 protocol inbound (backward). */
+    R2_SENT_SEIZE_ACK                           = 301,
+    /* Asking for calling party's category. */
+    R2_SENT_CATEGORY_REQUEST                    = 302,
+    /*! Asking for the next DNIS digit. */
+    R2_SENT_DNIS_REQUEST                        = 303,
+    /*! Asking for next ANI digit. */
+    R2_SENT_ANI_REQUEST                         = 304,
+    /*! Asking for change to group II. */
+    R2_SENT_CHANGE_TO_GROUP_II                  = 305,
+
+    /* Group B states for R2 protocol inbound (backward) */
+    /*! Sent numbers complete and start ANSWER. Charge. */
+    R2_SENT_ACCEPTED_CHARGE                     = 306,
+    /*! Sent numbers complete and start ANSWER. No charge. */
+    R2_SENT_ACCEPTED_NO_CHARGE                  = 307,
+    /*! Sent numbers complete. Route congested. */
+    R2_SENT_NETWORK_CONGESTION                  = 308,
+    /*! Sent numbers complete. Destination number unassigned. */
+    R2_SENT_UNASSIGNED_NUMBER                   = 309,
+    /*! Sent numbers complete. Destination number busy. */
+    R2_SENT_DEST_BUSY                           = 310,
+    /*! Sent numbers complete. Destination number out of order. */
+    R2_SENT_DEST_OUT_OF_ORDER                   = 311,
+
+    /* Group I states for R2 protocol outbound (forward) */
+    R2_SENT_CATEGORY                            = 401,
+    R2_SENT_DNIS                                = 402,
+    R2_SENT_ANI                                 = 403,
+    R2_SENT_END_OF_DNIS                         = 404,
+    R2_SENT_END_OF_ANI                          = 405,
+    R2_SENT_NOTHING                             = 406
+} r2_state_e;
 
 enum
 {
@@ -331,76 +323,40 @@ enum
     MFCR2_PROT_SAUDI_ARABIA,
     MFCR2_PROT_SINGAPORE,
     MFCR2_PROT_THAILAND,
-    MFCR2_PROT_VENEZUELA
+    MFCR2_PROT_VENEZUELA,
+    MFCR2_PROT_VIETNAM
 };
 
 /* This structure should contain everything to control an R2 channel. */
 struct mfcr2_signaling_state_s
 {
-    /*! \brief The current phase of the R2 protocol. */
-    int r2_phase;
-    /*! \brief The current compelled signalling state of the R2 protocol. */
-    int r2_state;
-    /*! \brief The total length of the DNIS string for outgoing calls. A progressively
-               incrementing count of the DNIS digits for incoming calls. */
-    int dnis_len;
-    /*! \brief The total length of the ANI string for outgoing calls. A progressively
-               incrementing count of the ANI digits for incoming calls. */
-    int ani_len;
-    /*! \brief The DNIS digits sent so far for outgoing calls. Not used for incoming calls. */
-    int dnis_cnt;
-    /*! \brief The ANI digits sent so far for outgoing calls. Not used for incoming calls. */
-    int ani_cnt;
-    /*! \brief The current CAS ABCD pattern being received. */
-    uint8_t current_abcd_in;
-    /*! \brief Timer used to ensure we do not seize to aggressively as the far end unblocks. */
-    int far_unblocking_timer;
-    /*! \brief Timer used to ensure we do not seize to aggressively after we unblock. */
-    int local_unblocking_timer;
-    /*! \brief The seize acknowledge timeout, used to detect failure at the remote end. */
-    int seize_ack_timer;
-    /*! \brief The release guard timeout. */
-    int release_guard_timer;
-    /*! \brief Optional minimum delay between sending the group B signal and
-        answering.
-    
-        If we answer too soon after sending the group B signal, the
-        outgoing side might see the answer before its tone detector has
-        found the group B signal. This timer allows some delay to be
-        introduced, to prevent such overlap conditions. */
-    int answer_guard_timer;
-    int wait_for_answer_timer;
-    int wait_for_group_b_signal_timer;
-    /*! \brief The MF tone on timer, at the outgoing side. */
-    int t1_timer;
-    /*! \brief The MF tone pulse timer, at the incoming side. This is only used
-               for variants which use non-compelled MF6 tone pulses to restart
-               activities after certain timeouts. */
-    int t1a_timer;
-    /*! \brief The MF tone off timer, at the outgoing side. */
-    int t2_timer;
-    /*! \brief The compelled cycle timer, at the incoming side. */
-    int t3_timer;
-    int far_blocked;
-    int local_blocked;
-    r2_mf_rx_state_t mf_rx_state;
-    tone_gen_state_t mf_gen_state;
-    int mf_rx_signal;
-    int mf_tx_signal;
-
     /* Configuration parameters, to adapt to different countries. */
     /*! \brief The number of DNIS digits we expect for incoming calls. Some national
                variants do not handle variable length dialed digit strings well. In
                those cases it is important to set this correctly. For other variants,
                a large number is OK. */
-    int dnis_num_rx_digits;
+    int dnis_max_rx_digits;
     /*! \brief The number of ANI digits we expect for incoming calls. */
     int ani_max_rx_digits;
-
+    /*! \brief TRUE to play progress tones for busy, NU, congested, etc. */
+    int play_progress_tones;
+    /*! \brief TRUE to play the disconnect tone after the local end has disconnected. */
+    int play_disconnect_tone;
+    /*! \brief TRUE to play the ringback tone while waiting for answer. */
+    int play_ringback_tone;
+    /*! \brief TRUE to use immediate accept (i.e. skipping the group B/II stage). */
+    int use_immediate_accept;
+    /*! \brief TRUE to get the ANI digits after the DNIS is complete. FALSE to get ANI after
+               the first DNIS digit. */
+    int get_ani_after_dnis;
     /*! \brief Forward tone max. on time, in milliseconds */
     int t1;
     /*! \brief Backward pulse on time (non-compelled), in milliseconds */
     int t1a;
+    /*! \brief Forward side max wait for the far end to timeout and recover, when
+               the end of digits is indicated by no reponse from our end. In
+               milliseconds. */
+    int t1b;
     /*! \brief Forward tone max. off time, in milliseconds */
     int t2;
     /*! \brief Inbound side compelled cycle max. time, in milliseconds */
@@ -419,7 +375,7 @@ struct mfcr2_signaling_state_s
     int max_seize_ack_wait;
     /*! \brief The maximum time we will for the called party to answer, after the
                call setup is complete. 60s is common practice */
-    int max_wait_for_answer;
+    int max_await_answer;
     /*! \brief The maximum time we will wait for the group B signal. 15s is common
                practice */
     int max_wait_for_group_b_signal;
@@ -430,25 +386,41 @@ struct mfcr2_signaling_state_s
                trunk is permitted to also make outgoing calls. */
     int outgoing_calls_allowed;
 
-    uint8_t fwd_abcd_idle;
+    /*! \brief Mask to pick out the R2 signaling bits from amongst the ABCD bits. */
+    uint8_t abcd_mask;
+    /*! \brief The value to use on transmission for the ABCD bits which are not the
+               R2 signaling bits. */
+    uint8_t base_abcd_bits;
+    /*! \brief The idle pattern for the ABCD bits */
+    uint8_t abcd_idle;
+    /*! \brief The blocking pattern for the ABCD bits */
+    uint8_t abcd_blocking;
+
+    /*! \brief The seize pattern for the ABCD bits */
     uint8_t fwd_abcd_seize;
+    /*! \brief The clear forward pattern for the ABCD bits */
     uint8_t fwd_abcd_clear_fwd;
 
-    uint8_t back_abcd_idle;
+    /*! \brief The seize acknowledge pattern for the ABCD bits */
     uint8_t back_abcd_seize_ack;
+    /*! \brief The answer pattern for the ABCD bits */
     uint8_t back_abcd_answer;
+    /*! \brief The clear back pattern for the ABCD bits */
     uint8_t back_abcd_clear_back;
-    uint8_t back_abcd_blocking;
+
+    uint8_t abcd_billing_pulse;
+    int billing_pulse_duration;
 
     /* Group I signals. */
     /*! \brief This is 0 (not supported) in some countries. */
     uint8_t group_i_end_of_DNIS;
+    /*! \brief This is 0 (not supported) in some countries. */
     uint8_t group_i_end_of_ANI;
     /*! \brief This is 0 (not supported) in some countries. */
     uint8_t group_i_end_of_ANI_restricted;
 
     /* Group II signals. */
-    uint8_t group_ii_xxx;
+    uint8_t group_ii[10];
 
     /* Group A signals. */
     /*! \brief Request the next digit of the DNIS
@@ -462,9 +434,12 @@ struct mfcr2_signaling_state_s
     uint8_t group_a_send_previous_DNIS_digit;
     /*! \brief This is the end of the transfer of numeric information. */
     uint8_t group_a_switch_to_group_ii;
+    /*! \brief Congestion has been detected. */
     uint8_t group_a_network_congestion;
+    /*! \brief Send the calling party category. */
     uint8_t group_a_send_calling_party_category;
-    uint8_t group_a_immediate_connect;
+    /*! \brief Immediately accept the call, without the group B/II step. */
+    uint8_t group_a_immediate_accept;
     /*! \brief Request the DNIS digit 2 digits back be repeated.
     
         This is an ITU specified feature, but a number of variants
@@ -485,11 +460,15 @@ struct mfcr2_signaling_state_s
     uint8_t group_a_repeat_all_DNIS_digits;
     /*! \brief The group A signal to request the next ANI digit. */
     uint8_t group_a_send_next_ANI_digit;
-    
+    /*! \brief The group A signal to switch to group C (only used in a
+        couple of countries). */
+    uint8_t group_a_send_calling_party_category_and_switch_to_group_c;    
+
     /* Group B signals. */
+    /*! \brief The group B signal to send SIT. */
     uint8_t group_b_send_special_information_tone;
     /*! \brief The group B signal to indicate the called line is busy. */
-    uint8_t group_b_line_busy;
+    uint8_t group_b_user_busy;
     /*! \brief The group B signal to indicate the called line cannot be reached, due to
                network congestion. */
     uint8_t group_b_network_congestion;
@@ -504,16 +483,103 @@ struct mfcr2_signaling_state_s
     /*! \brief The group B signal to indicate the called line is out of order. */
     uint8_t group_b_line_out_of_order;
     
+    /* Group C signals (only used in a couple of countries). */
+    uint8_t group_c_send_next_ANI_digit;
+    uint8_t group_c_repeat_all_DNIS_digits_and_switch_to_group_a;
+    uint8_t group_c_send_next_DNIS_digit_and_switch_to_group_a;
+    uint8_t group_c_send_previous_DNIS_digit_and_switch_to_group_a;
+    uint8_t group_c_switch_to_group_ii;
+
     /* Special things for some MFC/R2 variants */
     /*! \brief This flag determines is a timeout waiting for a DNIS digit on an
                incoming call should be treated as the end of the DNIS. */
     int group_b_on_DNIS_timeout;
 
+    /*! \brief The current call information. */
     uc_call_t *call;
     
     /*! \brief Alarms */
     int alarms;
 
+    /*! \brief The current phase of the R2 protocol. */
+    int r2_phase;
+    /*! \brief The current compelled signalling state of the R2 protocol. */
+    int r2_state;
+    /*! \brief A flag used to differentiate calling party category requests from
+               ANI digit requests. They use the same tone in many places. */
+    int calling_party_category_exchanged;
+    /*! \brief A flag which indicates all DNIS digits have been received. */
+    int dnis_complete;
+    /*! \brief The total length of the DNIS string for outgoing calls. A progressively
+               incrementing count of the DNIS digits for incoming calls. */
+    int dnis_len;
+    /*! \brief The total length of the ANI string for outgoing calls. A progressively
+               incrementing count of the ANI digits for incoming calls. */
+    int ani_len;
+    /*! \brief The DNIS digits sent so far for outgoing calls. Not used for incoming calls. */
+    int dnis_cnt;
+    /*! \brief The ANI digits sent so far for outgoing calls. Not used for incoming calls. */
+    int ani_cnt;
+    /*! \brief The ANI is restricted. */
+    int ani_restricted;
+    /*! \brief The current CAS ABCD pattern being received. */
+    uint8_t current_abcd_in;
+    /*! \brief Timer used to ensure we do not seize to aggressively as the far end unblocks. */
+    int far_unblocking_timer;
+    /*! \brief Timer used to ensure we do not seize to aggressively after we unblock. */
+    int local_unblocking_timer;
+    /*! \brief The seize acknowledge timeout, used to detect failure at the remote end. */
+    int seize_ack_timer;
+    /*! \brief The release guard timeout. */
+    int release_guard_timer;
+    /*! \brief Optional minimum delay between sending the group B signal and
+        answering.
+    
+        If we answer too soon after sending the group B signal, the
+        outgoing side might see the answer before its tone detector has
+        found the group B signal. This timer allows some delay to be
+        introduced, to prevent such overlap conditions. */
+    int answer_guard_timer;
+    /*! \brief A timer to handle the possibility that the called party never answers. */
+    int await_answer_timer;
+    /*! \brief A timer to handle the possibility that the group B signal never arrives. */
+    int wait_for_group_b_signal_timer;
+    /*! \brief Clear back persistence timer, used to differentiate between metering pulses
+        and real clear back conditions. */
+    int clear_back_persistence_timer;
+    /*! \brief Timer for the duration of a billing pulse. */
+    int billing_pulse_timer;
+    /*! \brief The MF tone on timer, at the outgoing side. */
+    int t1_timer;
+    /*! \brief The MF tone pulse timer, at the incoming side. This is only used
+               for variants which use non-compelled MF6 tone pulses to restart
+               activities after certain timeouts. */
+    int t1a_timer;
+    /*! \brief A backstop timer, to ensure things do not get stuck when
+               an outgoing call is relying on the other end moving on
+               when a timeout occurs. */
+    int t1b_timer;
+    /*! \brief The MF tone off timer, at the outgoing side. */
+    int t2_timer;
+    /*! \brief The compelled cycle timer, at the incoming side. */
+    int t3_timer;
+    /*! \brief TRUE is the far end is in the blocked state. */
+    int far_blocked;
+    /*! \brief TRUE is the local end is in the blocked state. */
+    int local_blocked;
+    /*! \brief The MF receiver context. */
+    r2_mf_rx_state_t mf_rx_state;
+    /*! \brief The MF transmitter context. */
+    r2_mf_tx_state_t mf_tx_state;
+    /*! \brief The current MF signal being received. Zero means no signal being received. */
+    int mf_rx_signal;
+    /*! \brief The current MF signal being sent. Zero means no signal being sent. */
+    int mf_tx_signal;
+    /*! \brief Supervisory tone generator context. */
+    super_tone_tx_state_t super_tone_tx_state;
+    /*! \brief Supervisory tone set. */
+    super_tone_set_t *super_tones;
+
     /*! \brief Debug audio receive log file handle. */
     int audio_rx_log;
     /*! \brief Debug audio transmit log file handle. */
